
=begin

== NAME

GDGraph.rb - Plotting Graph Library with Ruby/GD

Copyright (C) 2000  Noritsugu Nakamura


== EXAMPLE

=end

require 'GD'

module GDGraph
  DEBUG = false
  #DEBUG = true

  module Directional
	NOT_ASSIGNED = -1
	NO_DIRECTION = 0
	VERTICAL = 1
	HORIZONTAL = 2
  end

  module FileFormat
	PNG = 0
	GD = 1
	GD2 = 2
	JPEG = 3
  end

  module Color
	COLORS = {
	  'white'        => [255, 255, 255],
	  'black'        => [  0,   0,   0],
	  'red'          => [200,   0,   0],
	  'green'        => [  0, 175,   0],
	  'blue'         => [  0,   0, 200],
	  'orange'       => [250, 125,   0],
	  'yellow'       => [225, 225,   0],
	  'purple'       => [200,   0, 200],
	  'light_blue'   => [0,   125, 250],
	  'light_green'  => [125, 250,   0],
	  'light_purple' => [145,   0, 250],
	  'pink'         => [250,   0, 125],
	  'peach'        => [250, 125, 125],
	  'olive'        => [125, 125,   0],
	  'plum'         => [125,   0, 125],
	  'turquoise'    => [  0, 125, 125],
	  'mauve'        => [200, 125, 125],
	  'brown'        => [160,  80,   0],
	  'grey'         => [225, 225, 225]}

	COLORS_ary = [
	  'white',
	  'black',
	  'red',
	  'green',
	  'blue',
	  'orange',
	  'yellow',
	  'purple',
	  'light_blue',
	  'light_green',
	  'light_purple',
	  'pink',
	  'peach',
	  'olive',
	  'plum',
	  'turquoise',
	  'mauve',
	  'brown',
	  'grey',
	]

	WHITE = COLORS_ary.index('white')
	BLACK = COLORS_ary.index('black')
	RED = COLORS_ary.index('red')
	GREEN = COLORS_ary.index('green')
	BLUE = COLORS_ary.index('blue')
	ORANGE = COLORS_ary.index('orange')
	YELLOW = COLORS_ary.index('yellow')
	PURPLE = COLORS_ary.index('purple')
	LIGHT_BLUE = COLORS_ary.index('light_blue')
	LIGHT_GREEN = COLORS_ary.index('light_green')
	LIGHT_PURPLE = COLORS_ary.index('light_purple')
	PINK = COLORS_ary.index('pink')
	PEACH = COLORS_ary.index('peach')
	OLIVE = COLORS_ary.index('olive')
	PLUM = COLORS_ary.index('plum')
	TURQUOISE = COLORS_ary.index('turquoise')
	MAUVE = COLORS_ary.index('mauve')
	BROWN = COLORS_ary.index('brown')
	GREY = COLORS_ary.index('grey')

	def color(*arg)
	  case arg.size
	  when 1
		return COLORS_ary.index(arg[0])
	  when 2
		im, color = arg
		if color.type == String
		  return im.colorClosest(*COLORS[color])
		elsif color.type == Fixnum || color.type == Bignum
		  if color > COLORS_ary.size
			color = color % COLORS_ary.size
		  end
		  return im.colorClosest(*COLORS[COLORS_ary[color]])
		else
		  raise ArgumentError, "color(im, color) color must be String or Integer"
		end
	  else
		
	  end
	end
	module_function :color
  end

  module Marker
	Circle = 0
	FilledCircle = 1
	Triangle = 2
	Diamond = 3
	Square = 4
	Cross = 5

=begin
STAR
PENTAGON
INVERTED_TRIANGLE
BOWTIE
JIGGER
ASTERISK
CROSSHAIRS
SPINNING_WHEEL
CROSSHATCH
=end
	
#  	def initialize
#  	  @index = 0
#  	end

	def draw_marker(im, x, y, style, size, width, color)
	  case style
	  when Circle
		im.arc(x, y, size, size, 0, 360, GDGraph::Color.color(im, color))
	  when FilledCircle
		im.arc(x, y, size, size, 0, 360, GDGraph::Color.color(im, color))
		im.fill(x, y, GDGraph::Color.color(im, color))
	  when Triangle
		poly = GD::Polygon.new
		poly.addPt(x, y - size)
		poly.addPt(x + size * Math.cos(Math::PI/6), y + size * Math.sin(Math::PI/6))
		poly.addPt(x - size * Math.cos(Math::PI/6), y + size * Math.sin(Math::PI/6))
		im.polygon(poly, GDGraph::Color.color(im, color))
	  when Diamond
		s = 1.0 / Math.sqrt(2)
		poly = GD::Polygon.new
		poly.addPt(x, y + size / s)
		poly.addPt(x + size / s, y)
		poly.addPt(x, y - size / s)
		poly.addPt(x - size / s, y)
		im.polygon(poly, GDGraph::Color.color(im, color))
	  when Square
		im.rectangle(x - size / 2.0, y - size / 2.0, x + size / 2.0, y + size / 2.0, GDGraph::Color.color(im, color))
	  when Cross
		im.line(x - size / 2.0, y, x + size / 2.0, y, GDGraph::Color.color(im, color))
		im.line(x, y - size / 2.0, x, y + size / 2.0, GDGraph::Color.color(im, color))
	  end
	end
	module_function :draw_marker
  end

  module Font
  end

  class Label
  end

  class Data
	def initialize
	  @y_data = []
	  @start_x = 0
	  @interval = 1
	end

	def set_data(data)
	  if data.type != Array
		raise ArgumentError, "`Array' required"
	  end
	  @y_data = data.dup
	end
	
	attr_accessor :start_x, :interval
	attr_reader :y_data
#	protected :y_data
	public :y_data, :start_x, :interval
  end

  class LabeledData < Data
	def initialize(ary)
	  @labels = ary.dup
	  super()
	end

	attr_accessor :labels
#	attr_reader :y_data
  end

  class RegularData < Data
	def initialize(interval, start)
	  super()
	  @interval = interval
	  @start_x = start
	end
  end


  class AxisDrawer
	def initialize
	  @box_axis = 1

	  @line_width = 1
	  @line_style = 0
	  @line_color = 1

	  @font_name
	  @font_size
	end

	def draw_axis(im, component)
	  if @box_axis == 1
		if component.direction == Directional::VERTICAL
		  x1 = component.x_margin
		  x2 = component.width - component.x_margin
		  y1 = component.height - component.y_margin
		  y2 = component.y_margin
		  print "axis: (#{x1}, #{y1}) - (#{x1}, #{y2})\n" if DEBUG
		  im.rectangle(x1, y1, x2, y2, GDGraph::Color.color(im, @line_color))
		elsif component.direction == Directional::HORIZONTAL
		  x1 = component.y_margin
		  x2 = component.width - component.y_margin
		  y1 = component.x_margin
		  y2 = component.height - component.x_margin
		  print "axis: (#{x1}, #{y1}) - (#{x2}, #{y2})\n" if DEBUG
		  im.rectangle(x1, y1, x2, y2, GDGraph::Color.color(im, @line_color))
		else
		end
	  else
		if component.direction == Directional::VERTICAL
		  x1 = component.x_margin
		  x2 = component.width - component.x_margin
		  y = component.height - component.y_margin
		  print "axis: (#{x1}, #{y}) - (#{x1}, #{y})\n" if DEBUG
		  im.line(x1, y, x2, y, GDGraph::Color.color(im, @line_color))

		  x = component.x_margin
		  y1 = component.y_margin
		  y2 = component.height - component.y_margin
		  print "axis: (#{x}, #{y1}) - (#{x}, #{y2})\n" if DEBUG
		  im.line(x, y1, x, y2, GDGraph::Color.color(im, @line_color))

		  if component.exist_data?(1) == true
			x = component.width - component.x_margin
			y1 = component.y_margin
			y2 = component.height - component.y_margin
			print "axis: (#{x}, #{y1}) - (#{x}, #{y2})\n" if DEBUG
			im.line(x, y1, x, y2, GDGraph::Color.color(im, @line_color))
		  end
		elsif component.direction == Directional::HORIZONTAL
		  x = component.y_margin
		  y1 = component.y_margin
		  y2 = component.height - component.y_margin
		  print "axis: (#{x}, #{y1}) - (#{x}, #{y2})\n" if DEBUG
		  im.line(x, y1, x, y2, GDGraph::Color.color(im, @line_color))
		end
	  end
	end

	def draw_ticks(im, component)
	end
  end

  class Legend
	def initialize
	  
	end
  end


  class GraphComponent
	def initialize(width = 640, height = 480)
	  @graph = []
	  @x_num = 0

	  @y_min = []
	  @y_max = []
	  @y_gMin = []
	  @y_gMax = []

	  @width = width
	  @height = height
	  @direction = Directional::NOT_ASSIGNED
	  @interval = 0
	  @labels = []
	  @x_margin = 30
	  @y_margin = 20

	  # color
	  @bg_color

	  # pie
	  @pie_x0 = width * 0.5
	  @pie_y0 = height * 0.5
	  @pie_width = width * 0.8
	  @pie_height = width * 0.8
	  @pie_start_angle = 270
	end

	def set_axisdrawer(ad)
	  @axisdrawer = ad
	end

	def x_size
	  if @direction == Directional::VERTICAL
		return @width - @x_margin * 2
	  else
		return @height - @x_margin * 2
	  end	  
	end

	def y_size
	  if @direction == Directional::VERTICAL
		return @height - @y_margin * 2
	  else
		return @width - @y_margin * 2
	  end
	end

	def x_min=(x_min)
	  raise NotImplementError, ""
	  @x_min = x_min
	end

	def y_min=(y_min)
	  raise NotImplementError, ""
	  @y_min[0] = y_min
	end

	def composable(graph)
	  if (@direction != Directional::NOT_ASSIGNED) &&
		  (@direction != graph.direction)
		return false
	  end

	  if @graph.size == 0
		return true
	  end

	  if graph.data.type == LabeledData
		i = 0
		while (i < @labels.size) && (i < graph.data.labels.size)
		  if graph.data.labels[i] != @labels[i]
			return false
		  end
		  i += 1
		end
	  elsif graph.data.type == RegularData
		if graph.data.interval != @interval
		  return false
		end
	  else
	  end
	  return true
	end

	def set_x_minmax(graph)
	  data = graph.data
	  if @graph.size == 0
		@x_min = data.start_x
		@x_max = data.start_x + data.interval * data.y_data.size
	  else
		min = data.start_x
		max = data.start_x + data.interval * data.y_data.size
		if min < @x_min
		  @x_min = min
		end
		if max > @x_max
		  @x_max = max
		end
	  end
	end

	def set_y_minmax(graph, second_axis)
	  size = [0, 0]
	  @graph.each do |g, sa|
		size[0] += 1 if sa == 0
		size[1] += 1 if sa == 1
	  end

	  if size[second_axis] == 0
		@y_min[second_axis] = graph.y_min
		@y_max[second_axis] = graph.y_max
	  else
		if graph.y_min < @y_min[second_axis]
		  @y_min[second_axis] = graph.y_min
		end
		if graph.y_max > @y_max[second_axis]
		  @y_max[second_axis] = graph.y_max
		end
	  end

	  if @y_gMin[second_axis] == nil
		if graph.y_gMin != nil
		  @y_gMin[second_axis] = graph.y_gMin
		  @y_gMax[second_axis] = graph.y_gMax
		end
	  elsif graph.y_gMin != nil
		if graph.y_gMin < @y_gMin[second_axis]
		  @y_gMin[second_axis] = graph.y_gMin
		end
		if graph.y_gMax > @y_gMax[second_axis]
		  @y_gMax[second_axis] = graph.y_gMax
		end
	  end
	end

	# from Chart
	def set_y_scale(min, max)
	  # calculate good max_val
	  if max < -10
		tmp = -max
		order = (Math.log(tmp) / Math.log(10)).to_i
		mult = (tmp / (10 ** order)).to_i
		tmp = (mult - 1) * (10 ** order)
		max = -tmp
	  elsif max < 0
		max = 0
	  elsif max > 10
		order = (Math.log(max) / Math.log(10)).to_i
		mult = (max / (10 ** order)).to_i
		max = (mult + 1) * (10 ** order)
	  elsif max >= 0
		max = 10
	  end

	  # calculate good min_val
	  if min < -10
		tmp = -min
		order = (Math.log(tmp) / Math.log(10)).to_i
		mult = (tmp / (10 ** order)).to_i
		tmp = (mult + 1) * (10 ** order)
		min = -tmp
	  elsif min < 0
		min = -10
	  elsif min > 10
		order = (Math.log(min) / Math.log(10)).to_i
		mult = (min / (10 ** order)).to_i
		min = mult * (10 ** order)
	  elsif min >= 0
		min = 0
	  end

	  return min, max
	end

	def	merge_labels(graph)
	  if @labels.size == 0
		@labels = graph.data.labels
		return
	  end
	  if graph.data.labels.size <= @labels.size
		return
	  end
	  @labels = graph.data.labels
	end

	def add(graph, second_axis = 0)
	  if !composable(graph)
		raise ArgumentError, "do not match graph type"
	  end

	  @x_num = graph.size if graph.size > @x_num
	  print "@x_num: #{@x_num}\n" if DEBUG
	  set_x_minmax(graph)
	  set_y_minmax(graph, second_axis)
	  @y_min[second_axis], @y_max[second_axis] = set_y_scale(@y_min[second_axis], @y_max[second_axis])
	  if @y_gMin[second_axis] != nil
		@y_gMin[second_axis], @y_gMax[second_axis] = set_y_scale(@y_gMin[second_axis], @y_gMax[second_axis])
	  end
	  @graph.push([graph, second_axis])

	  if graph.data.type == LabeledData
		merge_labels(graph)
	  elsif graph.data.type == RegularData
		@interval = graph.data.interval
	  else
	  end

	  if @direction == Directional::NOT_ASSIGNED
		@direction = graph.direction
	  end
	end

	def draw_graph(im, component)
	  if @graph.size > 0
		@graph.each do |graph, second_axis|
		  graph.draw_data(component, im, second_axis)
		end
	  end
	end

	def out(file, format = FileFormat::PNG)
	  if file.type == String
		f = open(file, 'w')
	  elsif file.type == File
		f = file
	  else
		raise ArgumentError, "`String or File' required"
	  end

	  @im = GD::Image.new(@width, @height)
	  @im.colorAllocate(255,255,255)
	  for i in GDGraph::Color::COLORS.values
		@im.colorAllocate(*i)
	  end
	  #@im.line(50, 50, 150, 50, GDGraph::Color.color(@im, 'blue'))
	  
	  draw_graph(@im, self)
	  @axisdrawer.draw_axis(@im, self)
	  @axisdrawer.draw_ticks(@im, self)

	  if format == FileFormat::PNG
		@im.png f
	  else
	  end
	  f.close if file.type == String
	end

	def data_to_gd(y, second_axis, max = @y_max, min = @y_min)
	  if @direction == Directional::VERTICAL
		return @y_margin + y_size - y_size * (y - min[second_axis]) / (max[second_axis] - min[second_axis]).to_f
	  else
		return @y_margin + y_size * (y - min[second_axis]) / (max[second_axis] - min[second_axis]).to_f
	  end
	end

	def exist_data?(axis_num)
	  size = [0, 0]
	  @graph.each do |g, sa|
		return true if sa == axis_num
	  end
	  return false
	end

	attr_accessor :width, :height, :x_min, :x_max, :y_min, :y_max
	attr_accessor :pie_x0, :pie_y0, :pie_width, :pie_height, :pie_start_angle
	attr_reader :direction, :x_margin, :y_margin, :y_gMin, :y_gMax, :x_num
	protected :composable
	public :data_to_gd, :exist_data?
  end

  class SimpleGraph
	def initialize(*data)
#	  @direction = Directional::NOT_ASSIGNED
	  @x_min = 0
	  @x_max = 0
	  case data.size
	  when 0
	  when 1
		set_data(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end

	  @line_width = 1
	  @line_style = 0
	  @line_color = 1

	  @marker_style = GDGraph::Marker::Circle
#	  @marker_style = GDGraph::Marker::Cross
	  @marker_size = 10
	  @marker_color = 1

	  @bar_width = 0.4
	  @bar_color = GDGraph::Color::BLUE
	  @bar_filled = false
	end

	def subtotal(ary)
	  total = 0
	  for i in ary
		total += i
	  end
	  return total
	end

	def set_data(data)
#	  if data.type != Array
#		raise ArgumentError, "`Array' required"
#	  end
	  @data = data
	  d = data.y_data
	  @size = d.size
	  if d[0].type == Array
		@y_min = d[0][0]
	  else
		@y_min = d[0]
	  end
	  @y_max = 0
	  @maxNumberInGroup = 0
	  @total = 0
	  for i in d
		if i.type == Array
		  if i.size > @maxNumberInGroup
			@maxNumberInGroup = i.size
		  end
		  for j in i
			if @y_max < j
			  @y_max = j
			end
			if @y_min > j
			  @y_min = j
			end
		  end
		else
		  if @y_max < i
			@y_max = i
		  end
		  if @y_min > i
			@y_min = i
		  end
		  @total += i
		end
	  end
	  
	  if d[0].type == Array
		@y_gMax = subtotal(d[0])
		@y_gMin = subtotal(d[0])
		for i in d
		  total = subtotal(i)
		  if @y_gMax < total
			@y_gMax = total
		  end
		  if @y_gMin > total
			@y_gMin = total
		  end
		  @total += total
		end
	  end
	end

	attr_reader :data, :direction, :maxNumberInGroup, :total
	attr_reader :x_min, :x_max, :y_min, :y_max, :y_gMin, :y_gMax, :size
	attr_accessor :line_width, :line_style, :line_color
	attr_accessor :bar_width, :bar_color, :bar_filled
	attr_accessor :marker_style, :marker_size, :marker_color
	protected :subtotal
	public :direction, :data, :total
	public :line_width, :line_style, :line_color
	public :bar_width, :bar_color, :bar_filled
  end

  class VerticalBarGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::VERTICAL
	  @composable = false
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (2.0 * component.x_num)
	  half_width = gap * @bar_width
	  y0 = component.data_to_gd(0, second_axis)
	  for i in 0...@data.y_data.size
		if @data.y_data[0].type == Array
		  x = component.x_margin + gap * (1 + i * 2)
		  y = component.data_to_gd(@data.y_data[i][0], second_axis)
		  print "(", i, ", ", @data.y_data[i][0], ") --> ", "(", x, ", ", y0, ")", "(", x, ", ", y, ")\n" if DEBUG
		else
		  x = component.x_margin + gap * (1 + i * 2)
		  y = component.data_to_gd(@data.y_data[i], second_axis)
		  print "(", i, ", ", @data.y_data[i], ") --> ", "(", x, ", ", y0, ")", "(", x, ", ", y, ")\n" if DEBUG
		end
		if @bar_filled == true
		  im.filledRectangle(x - half_width, [y0, y].min, x + half_width, [y0, y].max, GDGraph::Color.color(im, @bar_color))
		  im.rectangle(x - half_width, y0, x + half_width, y, GDGraph::Color.color(im, @line_color))
		else
		  im.rectangle(x - half_width, y0, x + half_width, y, GDGraph::Color.color(im, @line_color))
		end
	  end
	end

	attr_reader :data, :composable
  end

  class HorizontalBarGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::HORIZONTAL
	  @composable = false
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (2.0 * component.x_num)
	  half_width = gap * @bar_width
	  x0 = component.data_to_gd(0, second_axis)
	  for i in 0...@data.y_data.size
		if @data.y_data[0].type == Array
		  x = component.data_to_gd(@data.y_data[i][0], second_axis)
		  y = component.x_margin + gap * (1 + i * 2)
		  print "(", i, ", ", @data.y_data[i][0], ") --> ", "(", x0, ", ", y, ")", "(", x, ", ", y, ")\n" if DEBUG
		else
		  x = component.data_to_gd(@data.y_data[i], second_axis)
		  y = component.x_margin + gap * (1 + i * 2)
		  print "(", i, ", ", @data.y_data[i], ") --> ", "(", x0, ", ", y, ")", "(", x, ", ", y, ")\n" if DEBUG
		end
		if @bar_filled == true
		  im.filledRectangle([x0, x].min, y - half_width, [x0, x].max, y + half_width, GDGraph::Color.color(im, @bar_color))
		  im.rectangle(x0, y - half_width, x, y + half_width, GDGraph::Color.color(im, @line_color))
		else
		  im.rectangle(x0, y - half_width, x, y + half_width, GDGraph::Color.color(im, @line_color))
		end
	  end
	end

	attr_reader :data, :composable
  end

  class VerticalPointsGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::VERTICAL
	  @composable = false
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (component.x_num - 1) # x_num != 1
	  if @data.y_data[0].type == Array
		for i in 0...@data.y_data[0].size
		  for j in 0...@data.y_data.size
			x = component.x_margin + gap * j
			y = component.data_to_gd(@data.y_data[j][i], second_axis)
			print "(", i, ", ", @data.y_data[j][i], ") --> ", "(", x, ", ", y, ")\n" if DEBUG
			GDGraph::Marker.draw_marker(im, x, y, @marker_style, @marker_size, 0, @marker_color)
		  end
		end
	  else
		for i in 0...@data.y_data.size
		  x = component.x_margin + gap * i
		  y = component.data_to_gd(@data.y_data[i], second_axis)
		  print "(", i, ", ", @data.y_data[i], ") --> ", "(", x, ", ", y, ")\n" if DEBUG
		  GDGraph::Marker.draw_marker(im, x, y, @marker_style, @marker_size, 0, @marker_color)
		end
	  end
	end

	attr_reader :data, :composable
  end

  class VerticalStackedBarGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::VERTICAL
	  @composable = false
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (2.0 * component.x_num)
	  half_width = gap * @bar_width
	  print "y_gMin: ", component.y_gMin, "\n" if DEBUG
	  print "y_gMax: ", component.y_gMax, "\n" if DEBUG
	  #component.y_gMax[0] = 100
	  if @data.y_data[0].type == Array
		for i in 0...@data.y_data.size
		  sum = 0
		  x = component.x_margin + gap * (1 + i * 2)
		  for j in 0...@data.y_data[i].size
			y0 = component.data_to_gd(sum, second_axis, component.y_gMax)
			y = component.data_to_gd(sum + @data.y_data[i][j], second_axis, component.y_gMax)
			sum += @data.y_data[i][j]
			print "(", i, ", ", @data.y_data[i][j], ") --> ", "(", x, ", ", y0, ")", "(", x, ", ", y, ")\n" if DEBUG
			if @bar_filled == true
			  im.filledRectangle(x - half_width, [y0, y].min, x + half_width, [y0, y].max, GDGraph::Color.color(im, j + 2))
			  im.rectangle(x - half_width, y0, x + half_width, y, GDGraph::Color.color(im, @line_color))
			else
			  im.rectangle(x - half_width, y0, x + half_width, y, GDGraph::Color.color(im, @line_color))
			end
		  end
		end
	  else
		
	  end
	end

	attr_reader :data, :composable
  end

  class VerticalLinesGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::VERTICAL
	  @composable = true
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (component.x_num - 1)
	  print "@line_width: #{@line_width}\n" if DEBUG
	  brush = GD::Image.new(@line_width, @line_width)
	  brush.colorAllocate(*GDGraph::Color::COLORS[GDGraph::Color::COLORS_ary[@line_color]])
	  im.setBrush(brush)
	  if @data.y_data[0].type == Array
		for i in 0...@data.y_data[0].size
		  for j in 1...@data.y_data.size
			x1 = component.x_margin + gap * (j - 1)
			y1 = component.data_to_gd(@data.y_data[j-1][i], second_axis)
			x2 = component.x_margin + gap * j
			y2 = component.data_to_gd(@data.y_data[j][i], second_axis)
			print "(i,j) = (", i, ", ", j, ")  (", j - 1, ", ", @data.y_data[j-1][i], ") - (", j, ", ", @data.y_data[j][i], ") --> ",
			  "(", x1, ", ", y1, ") - (", x2, ", ", y2, ")\n" if DEBUG
			#im.line(x1, y1, x2, y2, GDGraph::Color.color(im, @line_color))
			im.line(x1, y1, x2, y2, GD::Brushed)
		  end
		end
	  else
		for i in 1...@data.y_data.size
		  x1 = component.x_margin + gap * (i - 1)
		  y1 = component.data_to_gd(@data.y_data[i-1], second_axis)
		  x2 = component.x_margin + gap * i
		  y2 = component.data_to_gd(@data.y_data[i], second_axis)
		  print "(", i - 1, ", ", @data.y_data[i-1], ") (", i, ", ", @data.y_data[i], ") --> ", "(", x1, ", ", y1, ") ", "(", x2, ", ", y2, ")\n" if DEBUG
		  #im.line(x1, y1, x2, y2, GDGraph::Color.color(im, @line_color))
		  im.line(x1, y1, x2, y2, GD::Brushed)
		end
	  end
	end

	attr_reader :data, :composable
  end

  class VerticalLinesPointsGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::VERTICAL
	  @composable = true
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (component.x_num - 1)
	  print "@line_width: #{@line_width}\n" if DEBUG
	  brush = GD::Image.new(@line_width, @line_width)
	  brush.colorAllocate(*GDGraph::Color::COLORS[GDGraph::Color::COLORS_ary[@line_color]])
	  im.setBrush(brush)
	  if @data.y_data[0].type == Array
		for i in 0...@data.y_data[0].size
		  x = component.x_margin
		  y = component.data_to_gd(@data.y_data[0][i], second_axis)
		  GDGraph::Marker.draw_marker(im, x, y, @marker_style, @marker_size, 0, @marker_color)
		  for j in 1...@data.y_data.size
			x1 = component.x_margin + gap * (j - 1)
			y1 = component.data_to_gd(@data.y_data[j-1][i], second_axis)
			x2 = component.x_margin + gap * j
			y2 = component.data_to_gd(@data.y_data[j][i], second_axis)
			print "(i,j) = (", i, ", ", j, ")  (", j - 1, ", ", @data.y_data[j-1][i], ") - (", j, ", ", @data.y_data[j][i], ") --> ",
			  "(", x1, ", ", y1, ") - (", x2, ", ", y2, ")\n" if DEBUG
			#im.line(x1, y1, x2, y2, GDGraph::Color.color(im, @line_color))
			im.line(x1, y1, x2, y2, GD::Brushed)
			GDGraph::Marker.draw_marker(im, x2, y2, @marker_style, @marker_size, 0, @marker_color)
		  end
		end
	  else
		x = component.x_margin
		y = component.data_to_gd(@data.y_data[0], second_axis)
		GDGraph::Marker.draw_marker(im, x, y, @marker_style, @marker_size, 0, @marker_color)
		for i in 1...@data.y_data.size
		  x1 = component.x_margin + gap * (i - 1)
		  y1 = component.data_to_gd(@data.y_data[i-1], second_axis)
		  x2 = component.x_margin + gap * i
		  y2 = component.data_to_gd(@data.y_data[i], second_axis)
		  print "(", i - 1, ", ", @data.y_data[i-1], ") (", i, ", ", @data.y_data[i], ") --> ", "(", x1, ", ", y1, ") ", "(", x2, ", ", y2, ")\n" if DEBUG
		  #im.line(x1, y1, x2, y2, GDGraph::Color.color(im, @line_color))
		  im.line(x1, y1, x2, y2, GD::Brushed)
		  GDGraph::Marker.draw_marker(im, x2, y2, @marker_style, @marker_size, 0, @marker_color)
		end
	  end
	end

	attr_reader :data, :composable
  end

  class VerticalComparisonGraph < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::VERTICAL
	  @composable = false
	end
	
	def draw_data(component, im, second_axis)
	  gap = component.x_size / (2.0 * component.x_num)
	  gap_sub = gap / @maxNumberInGroup.to_f
	  half_width = gap_sub * @bar_width
	  y0 = component.data_to_gd(0, second_axis)
	  if @data.y_data[0].type == Array
		for i in 0...@data.y_data.size
		  x_main = component.x_margin + gap * (1 + i * 2)
		  for j in 0...@data.y_data[i].size
			x = x_main + half_width * ((j * 2) - @maxNumberInGroup + 1)
			y = component.data_to_gd(@data.y_data[i][j], second_axis)
			print "(", i, ", ", @data.y_data[i][j], ") --> ", "(", x, ", ", y0, ")", "(", x, ", ", y, ")\n" if DEBUG
			if @bar_filled == true
			  im.filledRectangle(x - half_width, [y0, y].min, x + half_width, [y0, y].max, GDGraph::Color.color(im, j + 2))
			  im.rectangle(x - half_width, y0, x + half_width, y, GDGraph::Color.color(im, @line_color))
			else
			  im.rectangle(x - half_width, y0, x + half_width, y, GDGraph::Color.color(im, @line_color))
			end
		  end
		end
	  else
		
	  end
	end

	attr_reader :data, :composable
  end

  class PieChart < SimpleGraph
	def initialize(*data)
	  case data.size
	  when 0
	  when 1
		super(data[0])
	  else
		raise ArgumentError, "0 or 1 arguments required"
	  end
	  @direction = Directional::NO_DIRECTION
	  @composable = true
	end
	
	def draw_data(component, im, second_axis)
	  if @data.y_data[0].type == Array
		ary = []
		for i in 0...@data.y_data.size
		  ary.push(@data.y_data[i][0])
		end
	  else
		ary = @data.y_data
	  end

	  total = 0
	  for i in ary
		total += i
	  end
	  print "total: ", total, "\n" if DEBUG
	  cx = component.pie_x0; cy = component.pie_y0
	  width = component.pie_x0; height = component.pie_y0
	  start = component.pie_start_angle
	  for i in 0...ary.size
		angle = (ary[i].to_f / total) * 360
		start_x = cx + (width / 2.0) * Math.cos(2 * Math::PI * start / 360)
		start_y = cy + (height / 2.0) * Math.sin(2 * Math::PI * start / 360)
		mid_x = cx + (width / (2.0 * 2)) * Math.cos(2 * Math::PI * (start + angle / 2.0) / 360)
		mid_y = cy + (height / (2.0 * 2)) * Math.sin(2 * Math::PI * (start + angle/ 2.0) / 360)
		end_x = cx + (width / 2.0) * Math.cos(2 * Math::PI * (start + angle) / 360)
		end_y = cy + (height / 2.0) * Math.sin(2 * Math::PI * (start + angle) / 360)
		print "start: ", start, "  end: ", start + angle, "\n" if DEBUG
		print "start_x: ", start_x, "  start_y: ", start_y, "\n" if DEBUG
		print "mid_x: ", mid_x, "  mid_y: ", mid_y, "\n" if DEBUG
		print "end_x: ", end_x, "  end_y: ", end_y, "\n" if DEBUG
		color = GDGraph::Color.color(im, i + 2)
		print "fill_color: ", color, "\n" if DEBUG
		im.arc(cx, cy, width, height, start, start + angle, GDGraph::Color.color(im, 'black'))
		im.line(cx, cy, start_x, start_y, GDGraph::Color.color(im, 'black'))
		im.line(cx, cy, end_x, end_y, GDGraph::Color.color(im, 'black'))
		im.fillToBorder(mid_x, mid_y, GDGraph::Color.color(im, 'black'), color)
		start += angle
	  end
	end

	attr_reader :data, :composable
  end
end
